
/* 

PiSpirit Rover Robot
PiSpirit_Base_Sketch_Rev01
Version 1.1 08/2016

Significant portions of this code written by
Dustin Soodak for Plum Geek LLC. Some portions
contributed by Kevin King.
Portions from other open source projects where noted.
This code is licensed under:
Creative Commons Attribution-ShareAlike 2.0 Generic (CC BY-SA 2.0)
https://creativecommons.org/licenses/by-sa/2.0/
Visit http://www.plumgeek.com for PiSpirit information.
Visit http://www.arduino.cc to learn about the Arduino.

*/

#include "RingoHardware.h"
#include <SPI.h>
#include <FastLED.h>

//These objects are used by the FastLED library
CRGB leds[31];        //array to hold FastLED pixel data
CRGBPalette16 currentPalette;
TBlendType    currentBlending;
#define DATA_PIN 7
#define NUM_LEDS 31

unsigned long timeStamp;

void setup(){
  HardwareBegin();                    //initialize Spirit's Arduino processor to work with his circuitry
  SetAllPixelsRGB(0,0,0);             //turn off all pixels before we get started
  PlayQuickChirp();                   //Play startup chirp and blink eyes

  PIC_ReadResetStatus();
  PIC_ReadPower();                    //read present supply voltage / supply current from PIC processor
  populate_generic_SPI_BufferOut(0);  //load outgoing buffer, ready for next SPI packet exchange
  SPI_Enable();                       //turn on the SPI port

  FastLED.addLeds<NEOPIXEL, DATA_PIN>(leds, NUM_LEDS);  //start FastLED
                                                        //FastLED Tutorial: https://github.com/FastLED/FastLED/wiki/Basic-usage
                                                        //FastLED Docs: http://fastled.io/docs/3.1/class_c_fast_l_e_d.html
  FastLED.clear();
  FastLED.show();
  LimePalette();          //pallete for LED's

  SetPixelRGB(0,0,0,20);  //indicate power on. RPi will usually change this pixel to Green once it is up and running.  
  SetPixelRGB(14,0,0,15); //turn on wing tips as part of power-on sequence
  SetPixelRGB(26,0,0,15); //turn on wing tips as part of power-on sequence

  timeStamp = millis();   //get starting timestamp

  //servoSpeed(80);
  gripStow(); //125
  servoSpeed(3000);

  //PIC_UART_AddNull = 1;
  //PIC_INT0_UART_Rx_Null = 1;
  //PIC_SetOpMode();

  PIC_SetComUARTMode(1);  //set UART to Bluetooth mode
  PIC_UART_Baud(9600);    //default baud for HC-06 Bluetooth module

}


void loop(){

  //PIC_UART_IntRxCount(2);
  //PIC_UART_IntRxChar('e');

  String outString;
  String inString = "56k";
  
  PIC_SendUART(inString);     //send the string to the PIC so it can sent it via UART
  delay(6);                   //short delay to let UART beging clocking out
  //delayMicroseconds(800);
  
  //PIC_ReadUARTBufferCount();
  //PIC_ReadStatus();
  
  //Serial.print(PIC_UARTRxThreshInt); Serial.print("\t"); 
  //Serial.print(PIC_UARTRxNullInt); Serial.print("\t"); 
  //Serial.println(PIC_UARTTxInProg); 

  // keep polling PIC to wait for Tx to no longer be in progress
  PIC_ReadStatus();
  while (PIC_UARTTxInProg){
    PIC_ReadStatus();
  }

  // Wait for tx to send fully
  //delay(20);  // extra delay to wait for response from other device
  
  // Read any response
  outString = PIC_ReadUART();
  Serial.print(outString);    //print received string to the serial monitor

  Serial.println();
  Serial.println();
  delay(500);   //extra delay before looping



  /*  //remove to re-enable SPI
  
  if (SPI_InProgress && (digitalRead(SPI_SS)==HIGH)){ //if SPI transaction has begun and SS line has returned high... (end of packet indication)

    SPI_Disable();    //disable the SPI port while we process the data

    if (computeSPIChecksum(1) && (SPI_BufferIn[9] != 0)){
      PIC_ReadPower();                                    //read present supply voltage / supply current from PIC processor
      //Serial.print(powerVoltage); Serial.print("\t"); Serial.println(powerCurrent); //uncomment for local debugging
      populate_generic_SPI_BufferOut(SPI_BufferIn[0]);    //load outgoing buffer, ready for next SPI packet exchange

      if (SPI_BufferIn[0] == 200){
          processPS3StickData();    //process and respond to passed stick & button data from SixAxis/PS3 controller paired to the RPi      
      }
      else if (SPI_BufferIn[0] == 201){  //direct drive tilt servo
          servoTilt(SPI_BufferIn[1]);
          Serial.print("tilt ");Serial.println(SPI_BufferIn[1]);
      }
      else if (SPI_BufferIn[0] == 202){  //direct drive pan servo
          servoPan(SPI_BufferIn[1]);
          Serial.print("pan ");Serial.println(SPI_BufferIn[1]);
      }
      else if (SPI_BufferIn[0] == 203){  //direct drive pan servo
          servoGrip(SPI_BufferIn[1]);
          Serial.print("grip ");Serial.println(SPI_BufferIn[1]);
      }
      else if (SPI_BufferIn[0] == 204){  //drive tilt and pan servos
          servoTilt(SPI_BufferIn[1]);
          servoPan(SPI_BufferIn[2]);
          Serial.print("tilt ");Serial.print(SPI_BufferIn[1]);Serial.print("\t");
          Serial.print("pan ");Serial.println(SPI_BufferIn[2]);
      }
      else if (SPI_BufferIn[0] == 210){  //drive motors
          int motor_left, motor_right;
          motor_left = (SPI_BufferIn[1]-128)*2;
          motor_right = (SPI_BufferIn[2]-128)*2;
          motors(motor_left,motor_right);
          Serial.print("motors ");Serial.print("\t");
          Serial.print(motor_left);Serial.print("\t");Serial.println(motor_right);
      }
      
      else if (SPI_BufferIn[0] == 161){  //kick off pre-set Arduino based sequence
          int motor_left, motor_right, low_thresh, high_thresh, run_time;
          motor_left = (SPI_BufferIn[1]-128)*2;
          motor_right = (SPI_BufferIn[2]-128)*2;
          low_thresh = (SPI_BufferIn[3]<<8)|SPI_BufferIn[4];     //create int from pair of passed SPI values
          high_thresh = (SPI_BufferIn[5]<<8)|SPI_BufferIn[6];    //create int from pair of passed SPI values
          run_time = (SPI_BufferIn[7]<<8)|SPI_BufferIn[8];       //create int from pair of passed SPI values
          Serial.print("driveToEdge ");Serial.print(motor_left);Serial.print("\t");Serial.print(motor_right);Serial.print("\t");
          Serial.print(low_thresh);Serial.print("\t");Serial.print(high_thresh);Serial.print("\t");Serial.println(run_time);
          driveToEdge(motor_left, motor_right, low_thresh, high_thresh, run_time);
      }

      else if (SPI_BufferIn[0] == 10){    
          int servo_speed = (SPI_BufferIn[1]<<8)|SPI_BufferIn[2]; //create int from pair of passed SPI values
          Serial.print("speed ");Serial.println(servo_speed);
          servoSpeed(servo_speed);  //set new servo speed in degr per second. Passing value of >= 255 sets max speed
      }
      else if (SPI_BufferIn[0] == 11){
          Serial.println("tilt");
          int servo_trim = SPI_BufferIn[1]-127;
          servoTilt_Trim(servo_trim);
      }
      else if (SPI_BufferIn[0] == 12){
          Serial.println("pan");
          int servo_trim = SPI_BufferIn[1]-127;
          servoPan_Trim(servo_trim);
      }
      else if (SPI_BufferIn[0] == 20){    
          int chirp = (SPI_BufferIn[1]<<8)|SPI_BufferIn[2]; //create int from pair of passed SPI values
          Serial.print("chirp ");Serial.println(chirp);
          PlayChirp(chirp);  //set new servo speed in degr per second. Passing value of >= 255 sets max speed
      }  
    }
    else{
      print_BufferIn();
      //Serial.print(SPI_BufferIn[0]); Serial.print("\t");Serial.print(SPI_BufferIn[1]); Serial.println("\t");
      Serial.print(computeSPIChecksum(1)); Serial.print("\t");Serial.print(SPI_BufferIn[9]); Serial.println("\t");
      //Serial.print(SPI_BufferIn[0]);Serial.print("\t");Serial.println(SPI_BufferIn[9]);
      Serial.println("***"); //indicates a bad packet was received, due to data error, or SPI speed set too high
    }

    SPI_Enable();               //re-enable the SPI port


  }  // end of if (SPI_InProgress), end of SPI handler

  */  //remove to re-enable SPI
} // end of loop() function

void printBits(byte myByte){
  for(byte mask = 0x80; mask; mask >>= 1){
    if(mask  & myByte)
        Serial.print('1');
    else
        Serial.print('0');
  }
}

void driveToEdge(int motor_left,int  motor_right,int  low_thresh,int  high_thresh,int  run_time){
  run_time = constrain(run_time, 0, 30000); //constrain runtime to between 0 and 30000 ms (30 seconds) 
  int reverse_time;
  timeStamp = millis();   //get starting timestamp  
  motors(motor_left,motor_right);   //start motors going
  boolean edgeDetected = 0;         //flag if we found an edge
  while (!edgeDetected){            //do this while edge is not detected
    if (millis() > timeStamp+run_time){ //if this goes longer than the expected runtime
      motors(0,0);  //stop motors
      Serial.println("timeout break");
      break;        //break out of the loop
    }
    PIC_ReadAllSurfaceSensors();      //read all surface sensors
    Serial.println("read sensors");
    // check if any surface sensor was below the passed low threshold value
    if (  (surfLeft1 < low_thresh) || (surfLeft0 < low_thresh) ||
          (surfRight0 < low_thresh) || (surfRight1 < low_thresh)){
            edgeDetected = 1;   //set flag that an edge was detected
            Serial.println("low_thresh");
          }
    // check if any surface sensor was above the passed high threshold value
    if (  (surfLeft1 > high_thresh) || (surfLeft0 > high_thresh) ||
          (surfRight0 > high_thresh) || (surfRight1 > high_thresh)){
            edgeDetected = 1;   //set flag that an edge was detected
            Serial.println("high_thresh");
          }
  } //end of while (!edgeDetected) loop
  Serial.println("end while");
  if (edgeDetected){  //this if will not run if the 'break' above ran due to timeout, so robot won't reverse on timeout
    Serial.println("edge was detected, reversing...");
    reverse_time = millis()-timeStamp;  //this tells us how long it took to find the edge in the above run
    
    motors(-motor_left,-motor_right);
    delay(250);
    motors(-motor_left,motor_right);
    delay(500);
    
    timeStamp = millis();               //get new timestamp
    motors(motor_left,motor_right);
    //motors(-motor_left,-motor_right);   //run motors in reverse of the passed values
    while(millis() < timeStamp+reverse_time){
      // do nothing (just let the motors drive while waiting for the timeout)
    }
    motors(0,0);    //stop motors. all done.
  }
} //end of driveToEdge() function

void print_BufferIn(void){
  for (int i=0; i<10; i++){
    Serial.print(SPI_BufferIn[i]);
    Serial.print("\t");
  }
  Serial.println("");
}


void populate_generic_SPI_BufferOut(byte response_register){
    SPI_BufferOut[0]= 0x00;                       //reserved for 8 alert flags
    SPI_BufferOut[1]= response_register;          //echo back response register
    SPI_BufferOut[2]= (powerVoltage >> 8) & 0xFF; //random(0,255);//232;  //data1
    SPI_BufferOut[3]= powerVoltage & 0xFF;        //random(0,255);//138;  //data2
    SPI_BufferOut[4]= (powerCurrent >> 8) & 0xFF; //random(0,255);//232;  //data3
    SPI_BufferOut[5]= powerCurrent & 0xFF;        //random(0,255);//138;  //data4
    SPI_BufferOut[6]= random(0,255);//122;        //data5
    SPI_BufferOut[7]= random(0,255);//174;        //data6
    SPI_BufferOut[8]= random(0,255);//183;        //data7
    SPI_BufferOut[9]= random(0,255);//229;        //data8
    computeSPIChecksum(0);      //compute checksum for outgoing SPI buffer & append checksum to end  
}

// FAST LED CODE

void LEDsFromPaletteColorsFrontBack( uint8_t colorIndex)
{
    uint8_t brightness = 100;
    
    for( int i = 3; i <= 14; i++) {
        leds[i] = ColorFromPalette( currentPalette, colorIndex, brightness, LINEARBLEND);
        colorIndex += 6;
    }
}

void LimePalette()
{
    CRGB green = CHSV( HUE_GREEN, 255, 255);
    CRGB yellow  = CHSV( HUE_YELLOW, 255, 255);
    CRGB yellowgreen  = CHSV( 75, 255, 255);
    CRGB black  = CRGB::Black;
    
    currentPalette = CRGBPalette16(
                                   green,  black,  yellowgreen,  black,
                                   yellowgreen,  black,  black,  black,
                                   green,  yellow,  yellowgreen,  green,
                                   yellowgreen,  black,  black,  black );
    currentBlending = LINEARBLEND;
}

void processPS3StickData(void){
  servoPan(SPI_BufferIn[3]);    //set head Pan servo based on value passed from RPi
  servoTilt(SPI_BufferIn[4]);   //set head Tilt servo based on value passed from RPi
  //set motor speeds based on value passed from RPi
  //value passed from RPi is 0-255, with 128 being center (no movement)
  //so we back out the 128 so center stick value results in motors(0,0) and max stick
  //values result in motors(-128,-128) to motors(128,128). Because our motors() function
  //goes from -255 to 255, we'll muliply this value by 2 to match the max range accepted
  //by our motors() function.
  motors((SPI_BufferIn[1]-128)*2, (SPI_BufferIn[2]-128)*2);
}

void SPI_Disable(void){       //disables the slave SPI port
  SPCR = 0x00;                // turn off SPI port
  SPI.detachInterrupt();      //turn off SPI interrupt
}

void SPI_Enable(void){        //enables the slave SPI port
  SPDR = 0x55;                //load outgoing SPI register for next send
  SPI_BufferPosition = 0;     //reset buffer position to the start
  SPI_InProgress = false;     //clear SPI in progress flag
  SPCR |= bit (SPE);          //turn on SPI in slave mode
  SPI.attachInterrupt();      //turn on interrupt for SPI port
}

void SPI_LoadOutgoingBuffer(void){   //load buffer for next SPI transaction
  //Load some random return data to the buffer for next time, to test the SPI connection
  SPI_BufferOut[1]= random(0,255);//123;  //response register
  SPI_BufferOut[2]= random(0,255);//232;  //data1
  SPI_BufferOut[3]= random(0,255);//138;  //data2
  SPI_BufferOut[4]= random(0,255);//143;  //data3
  SPI_BufferOut[5]= random(0,255);//50;  //data4
  SPI_BufferOut[6]= random(0,255);//122;  //data5
  SPI_BufferOut[7]= random(0,255);//174;  //data6
  SPI_BufferOut[8]= random(0,255);//183;  //data7
  SPI_BufferOut[9]= random(0,255);//229;  //data8
  computeSPIChecksum(0);      //compute checksum for outgoing SPI buffer & append checksum to end
}

// SPI interrupt routine
ISR (SPI_STC_vect)
{

  SPI_BufferIn[SPI_BufferPosition] = SPDR;        //copy incoming data byte from SPI port register
  SPDR = SPI_BufferOut[SPI_BufferPosition+1];     //copy next outgoing data byte to SPI port register
  
  SPI_InProgress = true;                          //set flag that a new packet is now in progress
  SPI_BufferPosition++;                           //set buffer index to next position of array

  if(SPI_BufferPosition > SPI_BUFFER_LENGTH-1){   //makes sure we don't over-run buffer and corrupt other data
    SPI_Disable();                                //turn off the SPI port so we don't get any more data
    SPI_BUFFER_LENGTH-1;                          //make sure subsequent bytes are written to end arry location
  }

}  // end of interrupt routine SPI_STC_vect




unsigned int integerValue=0;  // Max value is 65535
char incomingByte;

void getSerial(void){
  while(Serial.available() <= 0){
    //do nothing (wait for data from serial window)
  }
  if (Serial.available() > 0) {   // something came across serial
    integerValue = 0;         // throw away previous integerValue
    while(1) {            // force into a loop until 'n' is received
      incomingByte = Serial.read();
      if (incomingByte == '\n') break;   // exit the while(1), we're done receiving
      if (incomingByte == -1) continue;  // if no characters are in the buffer read() returns -1
      integerValue *= 10;  // shift left 1 decimal place
      // convert ASCII to integer, add, and shift left 1 decimal place
      integerValue = ((incomingByte - 48) + integerValue);
    }
    //regValue = integerValue;
    //Serial.print(lowByte(regValue));
    //Serial.print("\t");
    //Serial.println(highByte(regValue));   // Do something with the value
  }
}

void getTargetandValue(void){
  Serial.print("Target Register: ");
  while(Serial.available() <= 0){
    //do nothing (wait for data from serial window)
  }
  if (Serial.available() > 0) {   // something came across serial
    integerValue = 0;         // throw away previous integerValue
    while(1) {            // force into a loop until 'n' is received
      incomingByte = Serial.read();
      if (incomingByte == '\n') break;   // exit the while(1), we're done receiving
      if (incomingByte == -1) continue;  // if no characters are in the buffer read() returns -1
      integerValue *= 10;  // shift left 1 decimal place
      // convert ASCII to integer, add, and shift left 1 decimal place
      integerValue = ((incomingByte - 48) + integerValue);
    }
    regTarget = integerValue;
    Serial.println(regTarget);   // Do something with the value
  }

  Serial.print("Value: ");
  while(Serial.available() <= 0){
    //do nothing (wait for data from serial window)
  }
  if (Serial.available() > 0) {   // something came across serial
    integerValue = 0;         // throw away previous integerValue
    while(1) {            // force into a loop until 'n' is received
      incomingByte = Serial.read();
      if (incomingByte == '\n') break;   // exit the while(1), we're done receiving
      if (incomingByte == -1) continue;  // if no characters are in the buffer read() returns -1
      integerValue *= 10;  // shift left 1 decimal place
      // convert ASCII to integer, add, and shift left 1 decimal place
      integerValue = ((incomingByte - 48) + integerValue);
    }
    regValue = integerValue;
    Serial.print(lowByte(regValue));
    Serial.print("\t");
    Serial.println(highByte(regValue));   // Do something with the value
  }  
}


